#!/usr/bin/env python3

# Copyright (c) 2023, Natacha Porté
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import random
import sys

def inverse(l):
	result = [-1] * len(l)
	for i in range(len(l)):
		result[l[i]] = i
	return result

def name(i,s):
	if s == 27:
		return "ABC"[int(i/9)] + str(i%9)
	elif s == 54 or s == 52:
		return "J2" if i == 53 else \
		       "J1" if i == 52 else \
		       "ABCD"[int(i/13)] + "A234567890JQK"[i%13]
	elif s == 78:
		return "T" + str(i-56) if i >= 56 else \
		       "ABCD"[int(i/14)] + "A234567890VCDR"[i%14]
	else:
		return i

def names(l):
	return [name(i,len(l)) for i in l]

def mergesort(l,stacks):
#	print("starting mergesort({}, {})".format(l, stacks))
	result = -1
	if len(l) <= stacks:
		s = sorted(l)
		result = [[i, l.index(i)] for i in s]
	else:
		mark = max(l) + 100
		sublen = int(len(l) / stacks + 0.5)
		subres = [mergesort(l[(sublen*i):min(sublen*(i+1),len(l))], stacks) for i in range(stacks)]
		indices = [0] * stacks
		result = []
		while True:
			elts = [subres[i][indices[i]][0] \
			        if indices[i] < len(subres[i])
			        else mark
			        for i in range(stacks)]
			elt = min(elts)
			if elt == mark:
				break
			elti = elts.index(elt)
			assert(elt == subres[elti][indices[elti]][0])
			result.append([elt, elti] + subres[elti][indices[elti]][1:])
			indices[elti] += 1
#	print("returning mergesort({}, {}) -> {}".format(l, stacks, result))
	return result

size = int(sys.argv[1])
stacks = int(sys.argv[2]) if 2 < len(sys.argv) else 2
rounds = 0
while stacks ** rounds < size:
	rounds += 1
ref = [i for i in range(size)]
random.shuffle(ref)

print("Bulding {}".format(names(ref)))

permutation = inverse(ref)
state = [i for i in range(size)]

print("Inverse {}".format(names(permutation)))

m = mergesort(list(reversed(ref)) if rounds & 1 else ref, stacks)
state = [i for i in range(size)]
assert([c[0] for c in m] == state)
assert(len(m[0]) == rounds + 1)
print("<-- top    bottom -->")
for r in range(len(m[0])-1):
	print("Round {}:".format(r+1))
	n = [0] * stacks
	acc = [[] for i in range(stacks)]
	prev_stack = -1
	for c in state:
		cur_stack = m[c][-1-r]
		if r & 1 == 0:
			cur_stack = stacks - 1 - cur_stack
		if cur_stack < prev_stack:
			if max(n) > 0:
				print("".join(["   {}".format(i) for i in n]))
			n = [0] * stacks
		prev_stack = cur_stack
		acc[cur_stack].append(c)
		n[cur_stack] += 1
	if max(n) > 0:
		print("".join(["   {}".format(i) for i in n]))
	state = sum([list(reversed(a)) for a in acc], [])
	print(names(state))
